/*
 *            Copyright 2009-2023 The VOTCA Development Team
 *                       (http://www.votca.org)
 *
 *      Licensed under the Apache License, Version 2.0 (the "License")
 *
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *              http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#pragma once
#ifndef VOTCA_XTP_ACTIVEDENSITYMATRIX_H
#define VOTCA_XTP_ACTIVEDENSITYMATRIX_H
#include "logger.h"
#include "votca/xtp/orbitals.h"

namespace votca {
namespace xtp {
class ActiveDensityMatrix {
 public:
  ActiveDensityMatrix(Orbitals &orbitals, std::vector<Index> activeatoms,
                      double threshold = 0.4)
      : orbitals_(orbitals),
        activeatoms_(activeatoms),
        threshold_(threshold) {};
  std::array<Eigen::MatrixXd, 3> activedensitymatrix(
      const Eigen::MatrixXd &localized_mo_coeff);
  std::array<Eigen::MatrixXd, 3> compute_Dmat_A();

 private:
  Orbitals &orbitals_;
  std::vector<Index> activeatoms_;
  double threshold_;
};
}  // namespace xtp
}  // namespace votca
#endif