/*
 * Copyright 2009-2021 The VOTCA Development Team (http://www.votca.org)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// Standard includes
#include <iostream>

// Local VOTCA includes
#include "votca/csg/molecule.h"

namespace votca {
namespace csg {

using namespace std;

void Molecule::AddBead(Bead *bead, const string &name) {
  beads_.push_back(bead);
  bead_names_.push_back(name);
  beadmap_[name] = beads_.size() - 1;

  bead->setMoleculeId(id_);
}

Index Molecule::getBeadByName(const string &name) const {
  map<string, Index>::const_iterator iter = beadmap_.find(name);
  if (iter == beadmap_.end()) {
    std::cout << "cannot find: <" << name << "> in " << name_ << "\n";
    return -1;
  }
  return beadmap_.at(name);
}

}  // namespace csg
}  // namespace votca
